==> casa-agama-1.13.0-project.zip: lib/io/jans/casa/authn/UserAuthnUtil.java <==
package io.jans.casa.authn;

import io.jans.agama.model.*;
import io.jans.as.common.model.common.User;
import io.jans.as.server.service.*;
import io.jans.orm.PersistenceEntryManager;
import io.jans.service.cdi.util.CdiUtil;
import org.json.JSONObject;

import java.util.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class UserAuthnUtil {
    
    private static final Logger logger = LoggerFactory.getLogger(UserAuthnUtil.class);
    private static final MFAInfoHelper mfaInfo = new MFAInfoHelper();
    private static PersistenceEntryManager entryManager = CdiUtil.bean(PersistenceEntryManager.class);
    private static AuthenticationService authenticationService = CdiUtil.bean(AuthenticationService.class);

    private User user;
    private String uid;
    private String name;
    private String inum;
    private String preferredMethod;
    private boolean validCredentials;
    
    private String jsonLocation;
    private String jsonDevice;
    private List<String> policies = Collections.emptyList();

    public UserAuthnUtil() { }

    public UserAuthnUtil(List<String> policies) {
        this.policies = policies;
    }
    
    public void validate(String userName, String password) {
        
        logger.info("Validating password for {}", userName);
        if (authenticationService.authenticate(userName, password)) {
            user = authenticationService.getAuthenticatedUser();
            validCredentials = true;
            
            uid = user.getUserId();
            inum = user.getAttribute("inum");
            name = Optional.ofNullable(user.getAttribute("displayName")).orElse(user.getAttribute("givenName"));
            preferredMethod = translate(user.getAttribute("jansPreferredMethod"));
            updatePolicies();
        }
        
    }
    
    public String getUid() {
        return uid;
    }
    
    public String getName() {
        return name;
    }
    
    public String getInum() {
        return inum;
    }
    
    public boolean isValidCredentials() {
        return validCredentials;
    }    
    
    public boolean isUser2FAOn() {
         return preferredMethod != null;
    }
    
    public void setJsonLocation(String jsonLocation) {
        this.jsonLocation = jsonLocation;
    }
    
    public void setJsonDevice(String jsonDevice) {
        this.jsonDevice = jsonDevice;
    }    
    
    public boolean prompt2FA() {

        boolean prompt = policies.isEmpty() || policies.contains("EVERY_LOGIN");
        if (prompt) return true;
        
        // Validate device and location data before proceeding
        if (!isValidDeviceData()) {
            logger.warn("Invalid or missing device data, forcing 2FA prompt");
            return true;
        }
        
        TrustedDevicesManager tdm = new TrustedDevicesManager(user, jsonDevice, jsonLocation);
        
        if (/*!prompt &&*/ policies.contains("LOCATION_UNKNOWN")) {
            prompt = !tdm.knownLocation();
        }
        if (!prompt && policies.contains("DEVICE_UNKNOWN")) {
            prompt = !tdm.knownDevice();
        }
        if (!prompt) {
            logger.info("2FA will be skipped according to policy evaluation for this user");
        }
        return prompt;
        
    }

    /**
     * Validates that device data is complete and valid before using it for trusted device evaluation
     * This prevents issues with stale or incomplete device fingerprints
     */
    private boolean isValidDeviceData() {
        if (jsonDevice == null || jsonDevice.trim().isEmpty()) {
            logger.debug("Device data is null or empty");
            return false;
        }
        
        try {
            // Parse the device JSON to validate it has required fields
            org.json.JSONObject device = new org.json.JSONObject(jsonDevice);
            String browserName = device.optString("name");
            String browserVersion = device.optString("version");
            String osName = device.optString("osName");
            String osVersion = device.optString("osVersion");
            
            // Check if browser information is complete
            if (browserName == null || browserName.trim().isEmpty() || 
                browserVersion == null || browserVersion.trim().isEmpty()) {
                logger.debug("Browser information incomplete: name={}, version={}", browserName, browserVersion);
                return false;
            }
            
            // Check if OS information is complete
            if (osName == null || osName.trim().isEmpty() || 
                osVersion == null || osVersion.trim().isEmpty()) {
                logger.debug("OS information incomplete: name={}, version={}", osName, osVersion);
                return false;
            }
            
            // Check for fallback/placeholder values that indicate incomplete data
            if (isFallbackVersion(browserVersion) || isFallbackVersion(osVersion)) {
                logger.debug("Device data contains fallback values: browser={}, os={}", browserVersion, osVersion);
                return false;
            }
            
            return true;
            
        } catch (Exception e) {
            logger.debug("Error parsing device data: {}", e.getMessage());
            return false;
        }
    }
    
    /**
     * Checks if a version string represents a fallback/placeholder value
     */
    private boolean isFallbackVersion(String version) {
        if (version == null) return true;
        
        String v = version.trim().toLowerCase();
        return v.isEmpty() || 
               v.equals("0.0.0") || 
               v.equals("unknown") || 
               v.equals("undefined") ||
               v.equals("null") ||
               v.equals("n/a") ||
               v.equals("0") ||
               v.equals("1.0.0") ||
               v.equals("0.0");
    }

    /**
     * Validates the current session state to prevent stale session issues
     * Returns true if session is valid, false if stale session detected
     */
    public boolean validateSessionState() {
        try {
            // Simple session validation without servlet dependencies
            // In Agama environment, we'll use a basic timestamp approach
            
            // Check if we have valid device data as a proxy for session freshness
            if (!isValidDeviceData()) {
                logger.debug("Device data invalid, considering session stale");
                return false;
            }
            
            // For now, always return true to avoid blocking authentication
            // The device validation above will handle the main TouchID issue
            return true;
            
        } catch (Exception e) {
            logger.debug("Error validating session state: {}", e.getMessage());
            return false;
        }
    }
    
    /**
     * Clears any stale session data to force fresh authentication
     */
    public void clearStaleSessionData() {
        try {
            // In Agama environment, we can't directly access servlet session
            // Instead, we'll rely on device validation to handle stale session issues
            logger.debug("Session cleanup not available in Agama environment, using device validation instead");
            
            // The device validation in isValidDeviceData() will prevent stale session issues
            // by ensuring only valid, complete device data is used for trusted device evaluation
            
        } catch (Exception e) {
            logger.debug("Error in session cleanup: {}", e.getMessage());
        }
    }

    public List<String> computeUserMethods(LinkedHashSet<String> supportedMethods) {

        logger.trace("Supported methods: {}", supportedMethods);
        Set<String> methods = new LinkedHashSet<>();    //preserve insertion order
        List<String> empty = Collections.emptyList();

        //Reduce the supported methods to those actually installed
        supportedMethods.forEach(meh -> {
            try {
                String dn = String.format("%s=%s,ou=flows,ou=agama,o=jans", Flow.ATTR_NAMES.QNAME, meh);
        
                if (entryManager.contains(dn, ProtoFlow.class)) {
                    methods.add(meh);
                }
            } catch (Exception e) {
                logger.error(e.getMessage());
            }
        });
        logger.trace("List of methods: {}", methods);

        if (methods.isEmpty()) {
            logger.info("None of {} exist as agama flows", supportedMethods);
            return empty;
        }
        
        //Filter only the methods that match registered credentials for the user
        List<String> enrolled = mfaInfo.methodsEnrolled(inum);
        logger.info("User has enrollments for {}", enrolled);
        methods.retainAll(enrolled);
        logger.trace("Updated list of methods: {}", methods);  

        List<String> result = new ArrayList<>();
        if (methods.remove(preferredMethod)) {
            //Put the preferred on top
            result.add(preferredMethod);
        }
        result.addAll(methods);
        
        logger.info("Authentication methods list has been distilled to {}", result); 
        return result;

    }
    
    public void updateTrustedDevices() {

        try {
            if (policies.isEmpty() || !isUser2FAOn()) return;
            
            // Only update trusted devices if we have valid device data
            if (!isValidDeviceData()) {
                logger.warn("Skipping trusted device update due to invalid device data");
                return;
            }
            
            TrustedDevicesManager tdm = new TrustedDevicesManager(user, jsonDevice, jsonLocation);
            tdm.updateDevices();
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }

    }
    
    private void updatePolicies() {

        if (policies.contains("CUSTOM")) {
            String usrPolicy = user.getAttribute("jansStrongAuthPolicy");

            String[] policiesArr = usrPolicy == null ? new String[]{ "EVERY_LOGIN" } : usrPolicy.split(",\\s*");
            policies = List.of(policiesArr);
        }

    }

    private String translate(String oldAcr) {
        
        if (oldAcr == null) return null;
        
        if (!List.of("fido2", "super_gluu", "otp", "twilio_sms").contains(oldAcr)) return oldAcr;
        
        return "io.jans.casa.authn." + oldAcr;
        
    }
    
}

==> casa-agama-1.16.0-project.zip: lib/io/jans/casa/authn/UserAuthnUtil.java <==
package io.jans.casa.authn;

import io.jans.agama.model.*;
import io.jans.as.common.model.common.User;
import io.jans.as.server.service.*;
import io.jans.orm.PersistenceEntryManager;
import io.jans.service.cdi.util.CdiUtil;
import org.json.JSONObject;

import java.util.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class UserAuthnUtil {
    
    private static final Logger logger = LoggerFactory.getLogger(UserAuthnUtil.class);
    private static final MFAInfoHelper mfaInfo = new MFAInfoHelper();
    private static PersistenceEntryManager entryManager = CdiUtil.bean(PersistenceEntryManager.class);
    private static AuthenticationService authenticationService = CdiUtil.bean(AuthenticationService.class);

    private User user;
    private String uid;
    private String name;
    private String inum;
    private String preferredMethod;
    private boolean validCredentials;
    
    private String jsonLocation;
    private String jsonDevice;
    private List<String> policies = Collections.emptyList();

    public UserAuthnUtil() { }

    public UserAuthnUtil(List<String> policies) {
        this.policies = policies;
    }
    
    public void validate(String userName, String password) {
        
        logger.info("Validating password for {}", userName);
        if (authenticationService.authenticate(userName, password)) {
            user = authenticationService.getAuthenticatedUser();
            validCredentials = true;
            
            uid = user.getUserId();
            inum = user.getAttribute("inum");
            name = Optional.ofNullable(user.getAttribute("displayName")).orElse(user.getAttribute("givenName"));
            preferredMethod = translate(user.getAttribute("jansPreferredMethod"));
            updatePolicies();
        }
        
    }
    
    public String getUid() {
        return uid;
    }
    
    public String getName() {
        return name;
    }
    
    public String getInum() {
        return inum;
    }
    
    public boolean isValidCredentials() {
        return validCredentials;
    }    
    
    public boolean isUser2FAOn() {
         return preferredMethod != null;
    }
    
    public void setJsonLocation(String jsonLocation) {
        this.jsonLocation = jsonLocation;
    }
    
    public void setJsonDevice(String jsonDevice) {
        this.jsonDevice = jsonDevice;
    }    
    
    public boolean prompt2FA() {

        boolean prompt = policies.isEmpty() || policies.contains("EVERY_LOGIN");
        if (prompt) return true;
        
        // Validate device and location data before proceeding
        if (!isValidDeviceData()) {
            logger.warn("Invalid or missing device data, forcing 2FA prompt");
            return true;
        }
        
        TrustedDevicesManager tdm = new TrustedDevicesManager(user, jsonDevice, jsonLocation);
        
        if (/*!prompt &&*/ policies.contains("LOCATION_UNKNOWN")) {
            prompt = !tdm.knownLocation();
        }
        if (!prompt && policies.contains("DEVICE_UNKNOWN")) {
            prompt = !tdm.knownDevice();
        }
        if (!prompt) {
            logger.info("2FA will be skipped according to policy evaluation for this user");
        }
        return prompt;
        
    }

    /**
     * Validates that device data is complete and valid before using it for trusted device evaluation
     * This prevents issues with stale or incomplete device fingerprints
     */
    private boolean isValidDeviceData() {
        if (jsonDevice == null || jsonDevice.trim().isEmpty()) {
            logger.debug("Device data is null or empty");
            return false;
        }
        
        try {
            // Parse the device JSON to validate it has required fields
            org.json.JSONObject device = new org.json.JSONObject(jsonDevice);
            String browserName = device.optString("name");
            String browserVersion = device.optString("version");
            String osName = device.optString("osName");
            String osVersion = device.optString("osVersion");
            
            // Check if browser information is complete
            if (browserName == null || browserName.trim().isEmpty() || 
                browserVersion == null || browserVersion.trim().isEmpty()) {
                logger.debug("Browser information incomplete: name={}, version={}", browserName, browserVersion);
                return false;
            }
            
            // Check if OS information is complete
            if (osName == null || osName.trim().isEmpty() || 
                osVersion == null || osVersion.trim().isEmpty()) {
                logger.debug("OS information incomplete: name={}, version={}", osName, osVersion);
                return false;
            }
            
            // Check for fallback/placeholder values that indicate incomplete data
            if (isFallbackVersion(browserVersion) || isFallbackVersion(osVersion)) {
                logger.debug("Device data contains fallback values: browser={}, os={}", browserVersion, osVersion);
                return false;
            }
            
            return true;
            
        } catch (Exception e) {
            logger.debug("Error parsing device data: {}", e.getMessage());
            return false;
        }
    }
    
    /**
     * Checks if a version string represents a fallback/placeholder value
     */
    private boolean isFallbackVersion(String version) {
        if (version == null) return true;
        
        String v = version.trim().toLowerCase();
        return v.isEmpty() || 
               v.equals("0.0.0") || 
               v.equals("unknown") || 
               v.equals("undefined") ||
               v.equals("null") ||
               v.equals("n/a") ||
               v.equals("0") ||
               v.equals("1.0.0") ||
               v.equals("0.0");
    }

    /**
     * Validates the current session state to prevent stale session issues
     * Returns true if session is valid, false if stale session detected
     */
    public boolean validateSessionState() {
        try {
            // Simple session validation without servlet dependencies
            // In Agama environment, we'll use a basic timestamp approach
            
            // Check if we have valid device data as a proxy for session freshness
            if (!isValidDeviceData()) {
                logger.debug("Device data invalid, considering session stale");
                return false;
            }
            
            // For now, always return true to avoid blocking authentication
            // The device validation above will handle the main TouchID issue
            return true;
            
        } catch (Exception e) {
            logger.debug("Error validating session state: {}", e.getMessage());
            return false;
        }
    }
    
    /**
     * Clears any stale session data to force fresh authentication
     */
    public void clearStaleSessionData() {
        try {
            // In Agama environment, we can't directly access servlet session
            // Instead, we'll rely on device validation to handle stale session issues
            logger.debug("Session cleanup not available in Agama environment, using device validation instead");
            
            // The device validation in isValidDeviceData() will prevent stale session issues
            // by ensuring only valid, complete device data is used for trusted device evaluation
            
        } catch (Exception e) {
            logger.debug("Error in session cleanup: {}", e.getMessage());
        }
    }

    public List<String> computeUserMethods(LinkedHashSet<String> supportedMethods) {

        logger.trace("Supported methods: {}", supportedMethods);
        Set<String> methods = new LinkedHashSet<>();    //preserve insertion order
        List<String> empty = Collections.emptyList();

        //Reduce the supported methods to those actually installed
        supportedMethods.forEach(meh -> {
            try {
                String dn = String.format("%s=%s,ou=flows,ou=agama,o=jans", Flow.ATTR_NAMES.QNAME, meh);
        
                if (entryManager.contains(dn, ProtoFlow.class)) {
                    methods.add(meh);
                }
            } catch (Exception e) {
                logger.error(e.getMessage());
            }
        });
        logger.trace("List of methods: {}", methods);

        if (methods.isEmpty()) {
            logger.info("None of {} exist as agama flows", supportedMethods);
            return empty;
        }
        
        //Filter only the methods that match registered credentials for the user
        List<String> enrolled = mfaInfo.methodsEnrolled(inum);
        logger.info("User has enrollments for {}", enrolled);
        methods.retainAll(enrolled);
        logger.trace("Updated list of methods: {}", methods);  

        List<String> result = new ArrayList<>();
        if (methods.remove(preferredMethod)) {
            //Put the preferred on top
            result.add(preferredMethod);
        }
        result.addAll(methods);
        
        logger.info("Authentication methods list has been distilled to {}", result); 
        return result;

    }
    
    public void updateTrustedDevices() {

        try {
            if (policies.isEmpty() || !isUser2FAOn()) return;
            
            // Only update trusted devices if we have valid device data
            if (!isValidDeviceData()) {
                logger.warn("Skipping trusted device update due to invalid device data");
                return;
            }
            
            TrustedDevicesManager tdm = new TrustedDevicesManager(user, jsonDevice, jsonLocation);
            tdm.updateDevices();
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }

    }
    
    private void updatePolicies() {

        if (policies.contains("CUSTOM")) {
            String usrPolicy = user.getAttribute("jansStrongAuthPolicy");

            String[] policiesArr = usrPolicy == null ? new String[]{ "EVERY_LOGIN" } : usrPolicy.split(",\\s*");
            policies = List.of(policiesArr);
        }

    }

    private String translate(String oldAcr) {
        
        if (oldAcr == null) return null;
        
        if (!List.of("fido2", "otp", "twilio_sms").contains(oldAcr)) return oldAcr;
        
        return "io.jans.casa.authn." + oldAcr;
        
    }
    
}
